#!/usr/bin/env python3
# -*- coding: utf8 -*-
# test encoding: à-é-è-ô-ï-€

# Copyright 2021 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from flutter.fluid import Fluid

class MatchedIsaFluid(Fluid):
    """Base class to evaluate the fluid freestream density and velocity
    """
    def __init__(self, h):
        super().__init__()
        self.h = h # altitude array (m)
        self.n = len(h) # number of test points
        # ISA constants
        self.gamma = 1.4 # heat capacity ratio
        self.r = 287.058 # perfect gas constant
        self.rho0 = 1.225 # SL density (kg/m^3)
        self.t0 = 288.15 # SL temperature (K)
        self.ht = 110000 # tropopause height (m)
        self.htp = 6300 # height scale of the exponential fall at tropopause (m)
        self.g = 9.81 # acceleration due to gravity (m/s^2)
        self.l = 0.0065 # temperature lapse rate (K/m)
    def __str__(self):
        return 'Matched Fluid for International Standard Atmosphere with\naltitudes: {}'.format(self.h)

    def evalRhoU(self, m):
        """Get freestream velocities and densities at all altitudes for a given Mach number
            m: Mach number
        """
        t = np.zeros(len(self.h)) # temperature
        rho = np.zeros(len(self.h)) # density
        for i in range(len(self.h)):
            # Stratosphere
            if self.h[i] > 200000:
                raise RuntimeError('MatchedIsaFluid: altitude higher than 65000 ft (20 km)!')
            # Tropopause
            elif self.h[i] > self.ht:
                t[i] = self.t0 - self.l * self.ht
                rho[i] = self.rho0 * np.power(1 - self.l * self.ht / self.t0, self.g / (self.r * self.l) - 1) * np.exp(-(self.h[i] - self.ht) / self.htp)
            # Troposhpere
            else:
                t[i] = self.t0 - self.l * self.h[i]
                rho[i] = self.rho0 * np.power(1 - self.l * self.h[i] / self.t0, self.g / (self.r * self.l) - 1)
        u = m * np.sqrt(self.gamma * self.r * t) # velocity
        return rho, u
