#!/usr/bin/env python3
# -*- coding: utf8 -*-
# test encoding: à-é-è-ô-ï-€

# Copyright 2021 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import numpy as np
import scipy.linalg as spla
from flutter.solution import Solution

class Pk(Solution):
    """p-k solution method for the aeroelastic equations
    """
    def __init__(self, structural, aero, fluid):
        super().__init__(structural, aero, fluid)
    def __str__(self):
        return 'p-k solution method, with\n{}\n{}\n{}'.format(self.structural, self.aero, self.fluid)

    def run(self, tol=1e-4):
        """Compute frequency and damping ratio evolution with airspeed and find the flutter speed indices
            tol: tolerance for the frequency matching
        """
        # Start timer
        cpu = time.perf_counter()
        # Compute frequency and damping
        self.f = [np.zeros((self.fluid.n, self.structural.ns)) for _ in range(len(self.aero.machs))] # frequencies
        self.zeta = [np.zeros((self.fluid.n, self.structural.ns)) for _ in range(len(self.aero.machs))] # damping ratios
        print('Computing frequencies and damping ratios using p-k method...')
        print('{0:>5s} {1:>10s} {2:>5s} {3:>12s} {4:>12s}'.format('Mach', 'Airspeed', 'Mode', 'Frequency', 'Damping'))
        # Mach loop
        for i in range(len(self.aero.machs)):
            rho, u = self.fluid.evalRhoU(self.aero.machs[i]) # density and airspeed
            ks = np.zeros((self.fluid.n, self.structural.ns)) # reduced frequencies
            # Airspeed loop
            for j in range(self.fluid.n):
                # Structural mode loop
                for k in range(self.structural.ns):
                    ks[j, k] = self.structural.omega[k] * self.structural.lref / u[j] # reduced frequency (initial guess)
                    while True:
                        # Loads
                        Q = self.aero.compute(i, ks[j, k])
                        # Sorted eigenvalues
                        p, _ = spla.eig(self.structural.K - 0.5 * rho[j] * u[j]*u[j] * Q, -u[j]*u[j] / (self.structural.lref * self.structural.lref) * self.structural.M) # det(p^2 * u^2/l*M + K - 0.5*rho*u2*Q) = 0
                        p = np.sqrt(p)
                        p = np.where(np.imag(p) < 0 , -p , p) # change sign of p[i] if imag(p[i]) < 0
                        p = p[np.imag(p).argsort()]
                        # Update reduced frequency and terminate
                        error = abs(np.imag(p[k]) - ks[j, k])
                        ks[j, k] = np.imag(p[k])
                        if abs(error) < tol:
                            break
                    # Compute frequency and damping
                    self.f[i][j, k] = ks[j, k] * u[j] / (self.structural.lref * 2 * np.pi) # Hz
                    self.zeta[i][j, k] = -np.real(p[k]) / ks[j, k] # opposite of "true" damping, so that negative damping means instability
                    print('{0:5.3f} {1:10.3f} {2:5d} {3:12.6f} {4:12.6f}'.format(self.aero.machs[i], u[j], k, self.f[i][j, k], self.zeta[i][j, k]))
        # Find flutter speeds
        self.fidx = [[] for _ in range(len(self.aero.machs))]
        print('Finding flutter airspeeds...')
        print('{0:>5s} {1:>10s}'.format('Mach', 'Airspeeds'))
        for i in range(len(self.aero.machs)):
            diff = np.diff(np.sign(self.zeta[i]), axis=0) # consecutive differences in signs of damping...
            self.fidx[i] = np.unique(np.nonzero(diff)[0]) # ... nonzero elements indicate that sign has changed
            with np.printoptions(precision=1, suppress=True): print('{0:5.3f}  {1:}'.format(self.aero.machs[i], u[self.fidx[i]]))
        # Stop timer
        cpu = time.perf_counter() - cpu
        print('Computation done! Wall-clock time:', cpu, 's')

    def write(self, writer):
        """Save frequencies, damping ratios and flutter speeds to disk
            writer: type of writer to save the data
        """
        for i in range(len(self.aero.machs)):
            rho, u = self.fluid.evalRhoU(self.aero.machs[i]) # density and airspeed
            for j in range(self.structural.ns):
                # Frequencies and damping
                fname = 'f_zeta_mach{0:03d}_mode{1:03d}'.format(i, j)
                hdr = 'Frequencies and damping ratios for Mach = {0:5.3f}, Mode {1:03d}\n{2:>13s} {3:>15s} {4:>15s} {5:>15s}'.format(self.aero.machs[i], j, 'Density', 'Airspeed', 'Frequency', 'Damping')
                writer.write(fname, np.column_stack((rho, u, self.f[i][:, j], self.zeta[i][:,j])), hdr)
            # Flutter speed
            fname = 'uf_mach{0:03d}'.format(i)
            hdr = 'Flutter speeds for Mach = {0:5.3f}\n{1:>13s} {2:>15s}'.format(self.aero.machs[i], 'Density', 'Airspeed')
            writer.write(fname, np.column_stack((rho[self.fidx[i]], u[self.fidx[i]])), hdr)

    def plot(self):
        """Plot the frequencies and damping ratios vs airspeed
        """
        import matplotlib.pyplot as plt
        # Mach loop
        for i in range(len(self.aero.machs)):
            fig, axs = plt.subplots(2, 1)
            _, u = self.fluid.evalRhoU(self.aero.machs[i]) # airspeed
            for j in range(self.structural.ns):
                # Frequency
                axs[0].plot(u, self.f[i][:, j], label='Mode {0:3d}'.format(j), lw=2)
                axs[0].set_xlabel('Airspeed, $u$')
                axs[0].set_ylabel('Frequency, $f$ (Hz)')
                axs[0].legend()
                # Damping
                axs[1].plot(u, self.zeta[i][:, j], label='Mode {0:3d}'.format(j), lw=2)
                axs[1].set_xlabel('Airspeed, $u$')
                axs[1].set_ylabel('Damping, $\zeta$')
                axs[1].legend()
            axs[1].plot(u, [0.] * len(u), color='k', lw=1, ls='--') # zero-damping line
            fig.tight_layout()
            print('Saving figure to:', 'f_zeta_mach{0:03d}.png'.format(i), end='... ', flush=True)
            plt.savefig('f_zeta_mach{0:03d}.png'.format(i))
            print('done!')
