#!/usr/bin/env python3
# -*- coding: utf8 -*-
# test encoding: à-é-è-ô-ï-€

# Copyright 2021 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

class Structural:
    """Class for holding structural modal matrices and mode shapes
    """
    def __init__(self, m, l, o, a, p):
        self.ns = m # number of structural modes
        self.lref = l # reference length scale
        self.omega = o # natural (angular) frequencies
        self.amps = a # reference amplitude for mode shapes
        self.pfs = p # participation factor of each mode
    def __str__(self):
        return ('Structural Model with {:d} modes\n'.format(self.ns)
                +'reference length: {:5.2f}\n'.format(self.lref)
                +'natural frequencies: {}\n'.format(self.omega)
                +'reference amplitude: {}\n'.format(self.amps)
                +'participation factors: {}'.format(self.pfs)
                )

    def readMatrices(self, reader):
        """Read modal matrices from disk
        """
        self.M = reader.readMatrix('M') # mass matrix
        self.K = reader.readMatrix('K') # stiffness matrix
        if self.M.shape != (self.ns,self.ns):
            raise RuntimeError('Structural: modal mass matrix has shape {}, but {} was expected!'.format(self.M.shape, (self.ns,self.ns)))
        if self.K.shape != (self.ns,self.ns):
            raise RuntimeError('Structural: modal stiffness matrix has shape of {}, but {} was expected!'.format(self.M.shape, (self.ns,self.ns)))

    def readModes(self, reader, vname):
        """Read mode shapes from disk
        """
        self.modes = [None for _ in range(self.ns)] # None will be a np.array(n_points)
        # Read the modal displacements
        for i in range(self.ns):
           self.modes[i] = np.squeeze(reader.readValues('mode{:03d}'.format(i), [vname])[vname])
