#!/usr/bin/env python3
# -*- coding: utf8 -*-
# test encoding: à-é-è-ô-ï-€

# Copyright 2021 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from iops.reader import Reader
from msh.grid import Grid
from msh.node import Node
from msh.line import Line
import utils.general as gen

class TecplotReader(Reader):
    """Class for reading ASCII Tecplot data from disk
        caller: directory containing the files to be read
    """
    def __init__(self, caller):
        super().__init__(caller)
    def __str__(self):
        return 'Tecplot ASCII Reader with reference directory: "{:s}"'.format(self._dir)

    def __parse(self, f):
        """Parse the first zone of a Tecplot file
            f: file to be parsed
        """
        while True:
            line = f.readline().strip()
            # Read variables name
            if line.upper().startswith("VARIABLES"):
                vnames = self.__readVars(line) # variable names and number of coordinates
            # Read ONE zone
            elif line.upper().startswith("ZONE"):
                nnodes, ncells, nvertices = self.__readZoneHeader(line) # zone info (number of nodes, cells and type of cell)
                pointData, cellData = self.__readZoneData(f, nnodes, ncells, nvertices, len(vnames)) # zone data (nodal data and cell connectivity)
                return pointData, cellData, vnames
            # EOF
            elif not line:
                break
        raise RuntimeError('TecplotReader: end-of-file reached, but no data were read!')

    def __readVars(self, line):
        """Read the name of the variables
            line: line to be read
        """
        # Read variables names (expecting VARIABLES= "a", "b" ... OR "a""b" ...)
        line = line.split("=")[1].strip()
        if ',' in line:
            line = line.replace('"', '')
            line = [x for x in line.split(',')]
        else:
            line = line.strip('"')
            line = [x for x in line.split('""')]
        vnames = []
        for i in range(len(line)):
            line[i] = line[i].strip()
            vnames.append(line[i].strip('"'))
        return vnames

    def __readZoneHeader(self, line):
        """Read the information of the zone (number of nodes and cells and type of cell)
            line: line to be read
        """
        # Read (expecting ZONE  a=1, b=2 ...)
        line = line[5:]
        line = [x for x in line.split(',')]
        for i in range(len(line)):
            lin = line[i].split('=')
            # Number of nodes
            if lin[0].strip() == 'NODES':
                nnodes = int(lin[1])
            # Number of cells
            elif lin[0].strip() == 'ELEMENTS':
                ncells = int(lin[1])
            # Number of vertices per cell
            # a zone only contains one Tecplot cell type, but the nodes per cell can be duplicated (so that several type of cells are effectively in the same zone)
            elif lin[0].strip() == 'ZONETYPE':
                if lin[1].strip() == 'FELINESEG':
                    nvertices = 2
                elif lin[1].strip() == 'FETRIANGLE':
                    nvertices = 3
                elif lin[1].strip() == 'FEQUADRILATERAL':
                    nvertices = 4
                else:
                    raise RuntimeError('TecplotReader: Tecplot "ZONETYPE" not supported!')
        return nnodes, ncells, nvertices

    def __readZoneData(self, f, nnodes, ncells, nvertices, nvars):
        """Read nodal data and cell connectivity in Tecplot zone
            f: file being read
            nnodes: number of nodes
            ncells: number of cells
            nvertices: number of vertices per cell
            nvars: number of variables
        """
        # Read nodes data
        pointData = np.zeros((nnodes,nvars)) # pointData[i,j] will contain variable j for node i
        for i in range(nnodes):
            line = f.readline().strip().split('\t')
            if len(line) != nvars:
                raise RuntimeError('TecplotReader: wrong number of variables, expected {0:d} but read {1:d}!'.format(nvars, len(line)))
            for j in range(nvars):
                pointData[i,j] = float(line[j])
        # Read cells connectivity
        cellData = np.zeros((ncells,nvertices), dtype=int) # cellData[i,j] will contain node index j for cell i
        for i in range(ncells):
            line = f.readline().strip().split('\t')
            if len(line) != nvertices:
                raise RuntimeError('TecplotReader: Wrong number of vertices, expected {0:d} but read {1:d}!'.format(nvertices, len(line)))
            for j in range(nvertices):
                cellData[i,j] = int(line[j]) - 1 # Tecplot uses 1-based indexing
        return pointData, cellData

    def __createGrid(self, pointData, cellData, vnames):
        """Create a grid object
            pointData: nodal data
            cellData: cell connectivity
            vnames: names of variables contained in data
        """
        # Search for x, y, z
        cidx = []
        if 'x' in vnames:
            cidx.append(vnames.index('x'))
        elif 'X' in vnames:
            cidx.append(vnames.index('X'))
        else:
            raise RuntimeError('TecplotReader: variable "x" or "X" not found!')
        if 'y' in vnames:
            cidx.append(vnames.index('y'))
        elif 'Y' in vnames:
            cidx.append(vnames.index('Y'))
        else:
            raise RuntimeError('TecplotReader: variable "y" or "Y" not found!')
        if 'z' in vnames:
            cidx.append(vnames.index('z'))
        elif 'Z' in vnames:
            cidx.append(vnames.index('Z'))
        # Grid
        grid = Grid()
        # Nodes
        for i in range(pointData.shape[0]):
            coords = []
            for j in range(len(cidx)):
                coords.append(pointData[i,cidx[j]])
            grid.nodes.append(Node(i, coords))
        # Cells
        for i in range(cellData.shape[0]):
            nods = []
            for j in range(cellData.shape[1]):
                nods.append(grid.nodes[cellData[i,j]])
            nods = gen.removeDuplicates(nods) # remove possible duplicated entries
            if len(nods) == 2:
                grid.cells.append(Line(i, nods))
            elif len(nods) == 3:
                grid.cells.append(Tri(i, nods))
            elif len(nods) == 4:
                grid.cells.append(Quad(i, nods))
            else:
                raise RuntimeError('TecplotReader: unsupported cell type ({0:d} nodes)!'.format(len(nods)))
        return grid

    def __createValues(self, pointData, vnames, names):
        """Create a dict of values
            pointData: nodal data
            vnames: names of variables contained in data
            names: names of variables to extract
        """
        # Search for variables
        vidx = []
        for name in names:
            if name in vnames:
                vidx.append(vnames.index(name))
            else:
                raise RuntimeError('TecplotReader: variable "{0:s}" not found!'.format(name))
        # Variables
        values = {}
        for j in range(len(vidx)):
            values[names[j]] = pointData[:,vidx[j]]
        return values

    def readGrid(self, fname):
        """Read Tecplot file and return the grid
            fname: name of file contained in self._dir
        """
        # Open file
        fpth = super()._getFile(fname + '.dat') # return path to file
        print('Reading grid from:', fpth, end='... ', flush=True)
        with open(fpth, 'r') as f:
            pointData, cellData, vnames = self.__parse(f)
            grid = self.__createGrid(pointData, cellData, vnames)
        print('done!')
        return grid

    def readValues(self, fname, names):
        """Read Tecplot file and return the values at points
            fname: name of file contained in self._dir
            names: list of names of values to be read
        """
        # Open file
        fpth = super()._getFile(fname + '.dat') # return path to file
        print('Reading values from:', fpth, end='... ', flush=True)
        with open(fpth, 'r') as f:
            pointData, _, vnames = self.__parse(f)
            values = self.__createValues(pointData, vnames, names)
        print('done!')
        return values
