#!/usr/bin/env python3
# -*- coding: utf8 -*-
# test encoding: à-é-è-ô-ï-€

# Copyright 2021 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import vtk
import numpy as np
from iops.reader import Reader
from msh.grid import Grid
from msh.node import Node
from msh.line import Line
from msh.tri import Tri
from msh.quad import Quad

class VtkReader(Reader):
    """Class for reading data from disk using VTK
        caller: directory containing the files to be read
        fmt: file format (vtu, or vtk)
    """
    def __init__(self, caller, fmt):
        super().__init__(caller)
        # Create reader
        self.fmt = fmt
        if self.fmt == 'vtu':
            self.reader = vtk.vtkXMLUnstructuredGridReader()
        elif self.fmt == 'vtk':
            self.reader = vtk.vtkUnstructuredGridReader()
            self.reader.ReadAllScalarsOn()
            self.reader.ReadAllVectorsOn()
            self.reader.ReadAllTensorsOn()
            self.reader.ReadAllFieldsOn()
        else:
            raise Exception('Reader for format ' + self.fmt + ' not implemented!\n')
    def __str__(self):
        return 'VTK ({0:s} format) Reader with reference directory: "{1:s}"'.format(self.fmt, self._dir)

    def readGrid(self, fname):
        """Read VTK file and return the grid
            fname: name of file contained in self._dir
        """
        # Open file
        fpth = super()._getFile(fname + '.' + self.fmt) # return path to file
        print('Reading grid from:', fpth, end='... ', flush=True)
        self.reader.SetFileName(fpth)
        self.reader.Update()
        # Get grid data and create a grid object
        _grid = self.reader.GetOutput()
        grid = Grid()
        # Create nodes
        for i in range(_grid.GetNumberOfPoints()):
            grid.nodes.append(Node(i, _grid.GetPoint(i)))
        # Create cells
        for i in range(_grid.GetNumberOfCells()):
            _cell = _grid.GetCell(i)
            nods = []
            for j in range(_cell.GetNumberOfPoints()):
                nods.append(grid.nodes[_cell.GetPointId(j)])
            if (_cell.GetCellType() == vtk.VTK_LINE):
                grid.cells.append(Line(i, nods))
            elif (_cell.GetCellType() == vtk.VTK_TRIANGLE):
                grid.cells.append(Tri(i, nods))
            elif (_cell.GetCellType() == vtk.VTK_QUAD):
                grid.cells.append(Quad(i, nods))
            else:
                raise RuntimeError('VtkReader: unsupported cell type ({0:d} nodes)!'.format(len(nods)))
        print('done!')
        return grid

    def readValues(self, fname, names):
        """Read VTK file and return the values at points
            fname: name of file contained in self._dir
            names: list of names of values to be read
        """
        # Open file
        fpth = super()._getFile(fname + '.' + self.fmt) # return path to file
        print('Reading values from:', fpth, end='... ', flush=True)
        self.reader.SetFileName(fpth)
        self.reader.Update()
        # Get grid data...
        _grid = self.reader.GetOutput()
        values = {} # dictionary of values array
        for name in names:
            _values =  _grid.GetPointData().GetArray(name)
            values[name] = np.zeros((_values.GetNumberOfTuples(), _values.GetNumberOfComponents()))
            for i in range(values[name].shape[0]):
                for j in range(values[name].shape[1]):
                    values[name][i,j] = _values.GetTuple(i)[j]
        print('done!')
        return values
