#!/usr/bin/env python3
# -*- coding: utf8 -*-
# test encoding: à-é-è-ô-ï-€

# Copyright 2021 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import numpy as np
from iops import asciireader as ardr
from iops import tecplotreader as trdr
from iops import asciiwriter as awrt
from flutter import pk
from flutter import unmatchedfluid as flu
from flutter import structural as sa
from flutter import aero as aa
from utils import testing as tst

def main():
    """Test case for the 2D isogai airfoil
       from Hüseyin Güner thesis, chapter 4.2, pp71 (http://hdl.handle.net/2268/245578)
    """
    # Geometric and structural parameters
    ndim = 2 # dimension
    c = 1 # airfoil chord
    mu = 60 # mass ratio
    nmodes = 2 # number of structural modes
    omegan = np.array([100, 100]) # wind-off natural angular frequencies (plunge, pitch)
    amp = [0.1*c, 1*np.pi/180] # amplitude of modes
    pf = [1, 1] # modal participation factor
    # Flight and flutter parameters
    rho = 0.394353 # density
    machs = [0.800, 0.825, 0.850, 0.875] # Mach numbers
    krefs = [0.1, 0.3] # reference reduced frequencies
    ui = np.arange(0.3, 3.1, 0.1) # speed index

    # I/O
    matreadr = ardr.AsciiReader(os.path.join(os.path.dirname(__file__), 'data', os.path.basename(__file__)[:-3], 'structural')) # matrices
    modreadr = trdr.TecplotReader(os.path.join(os.path.dirname(__file__), 'data', os.path.basename(__file__)[:-3], 'structural')) # modes
    prsreadr = trdr.TecplotReader(os.path.join(os.path.dirname(__file__), 'data', os.path.basename(__file__)[:-3], 'aero')) # pressure
    writer = awrt.AsciiWriter()

    # Structural model
    struct = sa.Structural(nmodes, 0.5*c, omegan, amp, pf)
    struct.readMatrices(matreadr) # mass and stiffness matrices
    struct.readModes(modreadr, 'dy') # vertical modal displacement

    # Aerodynamic model
    fluid = flu.UnmatchedFluid(rho, ui, 0.5*c, struct.omega[1], mu) # density and airspeed
    aero = aa.Aero(machs, krefs)
    aero.fromTime(prsreadr, 'Pressure_Coefficient', ndim, struct) # modal loads matrix from time instances

    # Flutter solution using p-k method
    sol = pk.Pk(struct, aero, fluid)
    sol.run() # compute frequencies and damping and find flutter speed
    sol.write(writer) # save to disk
    if parse().fig:
        sol.plot() # display results

    # Testing
    tests = tst.Tests()
    tests.add(tst.Test('Flutter speed index (Mach 0.800)', ui[sol.fidx[0]], 0.8, 1e-2))
    tests.add(tst.Test('Flutter speed index (Mach 0.825)', ui[sol.fidx[1][0]], 0.6, 1e-2)) # 2 fluttter speeds due to numerical errors (mode 1 has negative damping at zero speed)
    tests.add(tst.Test('Flutter speed index (Mach 0.850)', ui[sol.fidx[2]], 1.3, 1e-2))
    tests.add(tst.Test('Flutter speed index (Mach 0.875)', ui[sol.fidx[3]], 2.4, 1e-2))
    tests.run()

if __name__ == "__main__":
    main()
